import logging

from crownstone_core.Exceptions import CrownstoneBleException
from crownstone_core.util.EncryptionHandler import EncryptionHandler

LAST_PACKET_INDEX = 0xFF

_LOGGER = logging.getLogger(__name__)

class NotificationDelegate:
    """
    Merges notifications and decrypts the merged data.
    The decrypted data is then placed in the "result" variable.
    """

    def __init__(self, callback, settings):
        self.callback = callback
        self.previousPart = -1 # Start at -1, so that we can check if received part > previous part
        self.dataCollected = []
        self.result = None
        self.settings = settings

    def handleNotification(self, uuid, data):
        self.merge(data)

    def merge(self, data):
        part = data[0]

        if self.result is not None:
            _LOGGER.debug(f"Last part already received, ignoring this part.")
            return

        # Ignore the case where we receive the same part twice.
        if part == self.previousPart:
            _LOGGER.debug(f"Already received part {part}, ignoring this part.")
            return

        # Check the part number.
        if part != LAST_PACKET_INDEX and part != self.previousPart + 1:
            _LOGGER.debug(f"Receive part {part}, expected part {self.previousPart + 1}")
            self.reset()
            return
        self.previousPart = part

        self.dataCollected += data[1:]
        _LOGGER.debug(f"Received part {part}")

        if data[0] == LAST_PACKET_INDEX:
            _LOGGER.debug(f"Received last part. Merged data: {self.dataCollected}")
            result = self.checkPayload()
            self.reset()
            self.result = result
            _LOGGER.debug(f"Result: {result}")
            if self.callback is not None:
                self.callback()

    def checkPayload(self):
        try:
            return EncryptionHandler.decrypt(self.dataCollected, self.settings)
        except CrownstoneBleException as err:
            _LOGGER.debug(f"Failed to decrypt: {err.message}")

    def reset(self):
        self.previousPart = -1
        self.dataCollected = []
        self.result = None



import asyncio
import logging

import bleak.exc
from bleak import BleakClient, BleakScanner

from crownstone_core.Exceptions import CrownstoneBleException, CrownstoneError
from crownstone_core.core.modules.EncryptionSettings import EncryptionSettings
from crownstone_core.protocol.BluenetTypes import ProcessType
from crownstone_core.util.EncryptionHandler import EncryptionHandler

from crownstone_ble.Exceptions import BleError
from crownstone_ble.core.BleEventBus import BleEventBus

from crownstone_ble.core.bluetooth_delegates.BleakScanDelegate import BleakScanDelegate
from crownstone_ble.core.bluetooth_delegates.NotificationDelegate import NotificationDelegate
from crownstone_ble.core.modules.Validator import Validator
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics

_LOGGER = logging.getLogger(__name__)

CCCD_UUID = 0x2902


class ActiveClient:

    def __init__(self, address, cleanupCallback, bleAdapterAddress):
        self.address = address
        if bleAdapterAddress is None:
            self.client = BleakClient(address)
        else:
            self.client = BleakClient(address, adapter=bleAdapterAddress)
        self.cleanupCallback = cleanupCallback
        self.client.set_disconnected_callback(self.forcedDisconnect)

        # Dict with service UUID as key, handle as value.
        self.services = {}

        # Dict with characteristic UUID as key, handle as value.
        self.characteristics = {}

        # Current callbacks for notifications, in the form of: def handleNotification(uuid: str, data)
        # Characteristic UUID is key, callback is value.
        self.notificationCallbacks = {}

        # Notifications we subscribed to.
        # Handle as key, UUID as value.
        self.notificationSubscriptions = {}

    def forcedDisconnect(self, data):
        BleEventBus.emit(SystemBleTopics.forcedDisconnect, self.address)
        self.cleanupCallback()

    async def isConnected(self):
        return await self.client.is_connected()

    async def subscribeNotifications(self, characteristicUuid: str, callback):
        _LOGGER.debug(f"register callback for notifications to uuid={characteristicUuid}")
        if characteristicUuid in self.notificationCallbacks:
            _LOGGER.error(f"There is already a callback registered for {characteristicUuid}")

        if characteristicUuid not in self.notificationSubscriptions.values():
            # handle = self.characteristics.get(uuid, None)
            # if handle is not None:
            _LOGGER.debug(f"subscribe to uuid={characteristicUuid}")
            handle = self.characteristics[characteristicUuid]
            await self.client.start_notify(characteristicUuid, self._resultNotificationHandler)
            self.notificationSubscriptions[handle] = characteristicUuid
        self.notificationCallbacks[characteristicUuid] = callback

    def unsubscribeNotifications(self, characteristicUuid: str):
        _LOGGER.debug(f"remove callback for notifications to uuid={characteristicUuid}")
        self.notificationCallbacks.pop(characteristicUuid, None)

    def _resultNotificationHandler(self, characteristicHandle, data):
        uuid = self.notificationSubscriptions.get(characteristicHandle, None)
        if uuid is None:
            _LOGGER.error(f"UUID not found for handle {characteristicHandle}")
        callback = self.notificationCallbacks.get(uuid, None)
        if callback is not None:
            callback(uuid, data)



class BleHandler:

    def __init__(self, settings: EncryptionSettings, bleAdapterAddress: str=None):
        # bleAdapterAddress is the MAC address of the adapter you want to use.

        self.settings = settings
        self.bleAdapterAddress = bleAdapterAddress

        # Connection
        self.activeClient: ActiveClient or None = None

        # Scanning
        self.scanner = BleakScanner(adapter=bleAdapterAddress)
        self.scanningActive = False
        self.scanAborted = False
        scanDelegate = BleakScanDelegate(self.settings)
        self.scanner.register_detection_callback(scanDelegate.handleDiscovery)

        # Event bus
        self.subscriptionIds = []
        self.validator = Validator()
        self.subscriptionIds.append(BleEventBus.subscribe(SystemBleTopics.abortScanning, lambda x: self.abortScan()))

        # To be moved to active client or notification handler.
        self.notificationLoopActive = False


    async def shutDown(self):
        for subscriptionId in self.subscriptionIds:
            BleEventBus.unsubscribe(subscriptionId)
        await self.disconnect()
        await self.stopScanning()


    async def is_connected_guard(self):
        connected = await self.is_connected()
        if not connected:
            _LOGGER.debug(f"Could not perform action since the client is not connected!.")
            raise CrownstoneBleException("Not connected.")


    async def is_connected(self, address = None) -> bool:
        """
        Check if connected to a BLE device.
        :param address: When not None, check if connected to given address.
        :returns:       True when connected.
        """
        if self.activeClient is None:
            return False
        if address is not None:
            if self.activeClient.address.lower() != address.lower():
                return False
        connected = await self.activeClient.client.is_connected()
        if connected:
            return True


    def resetClient(self):
        self.activeClient = None


    async def connect(self, address, timeout: int = 5, attempts: int = 3) -> bool:
        # Always clean up cached setup key for a new connection.
        self.settings.exitSetup()

        connected = await self.is_connected(address)
        if connected:
            _LOGGER.info("Already connected")
            return

        # TODO: Check if activeClient is already set.
        self.activeClient = ActiveClient(address, lambda: self.resetClient(), self.bleAdapterAddress)

        _LOGGER.info(f"Connecting to {address}")
        connected = False
        for i in range(0, attempts):
            connected = await self.connectAttempt(timeout)
            if connected:
                break
        if not connected:
            raise CrownstoneBleException(CrownstoneError.CONNECTION_FAILED)

        _LOGGER.info(f"Connected")
        serviceSet = await self.activeClient.client.get_services()
        self.activeClient.services = {}
        self.activeClient.characteristics = {}
        for key, service in serviceSet.services.items():
            self.activeClient.services[service.uuid] = key
        for key, characteristic in serviceSet.characteristics.items():
            self.activeClient.characteristics[characteristic.uuid] = characteristic.handle

        self.activeClient.notificationCallbacks = {}
        self.activeClient.notificationSubscriptions = {}

        return connected

    async def connectAttempt(self, timeout: int) -> bool:
        # this can throw an error when the connection fails.
        # these BleakErrors are nicely human readable.
        # TODO: document/convert these errors.
        try:
            _LOGGER.debug(f"Connecting..")
            connected = await self.activeClient.client.connect(timeout = timeout)
        except bleak.BleakError as err:
            _LOGGER.info(f"Failed to connect: {err}")
            connected = False
        except TimeoutError as err:
            _LOGGER.info(f"Failed to connect: {err}")
            connected = False
        return connected


    async def disconnect(self):
        if self.activeClient is not None:
            self.settings.exitSetup()
            await self.activeClient.client.disconnect()
            self.activeClient = None


    async def waitForPeripheralToDisconnect(self, timeout: int = 10):
        if self.activeClient is not None:
            if await self.activeClient.isConnected():
                waiting = True
                def disconnectListener(data):
                    nonlocal waiting
                    waiting = False

                listenerId = BleEventBus.subscribe(SystemBleTopics.forcedDisconnect, disconnectListener)

                timer = 0
                while waiting and timer < 10:
                    await asyncio.sleep(0.1)
                    timer += 0.1

                self.settings.exitSetup()
                BleEventBus.unsubscribe(listenerId)

                self.activeClient = None


    async def scan(self, duration=3):
        _LOGGER.debug(f"scan duration={duration}")
        await self.startScanning()
        while duration > 0 and self.scanAborted == False:
            await asyncio.sleep(0.1)
            duration -= 0.1
        await self.stopScanning()


    async def startScanning(self):
        _LOGGER.debug(f"startScanning scanningActive={self.scanningActive}")
        if not self.scanningActive:
            self.scanAborted = False
            self.scanningActive = True
            await self.scanner.start()


    async def stopScanning(self):
        _LOGGER.debug(f"stopScanning scanningActive={self.scanningActive}")
        if self.scanningActive:
            self.scanningActive = False
            self.scanAborted = False
            await self.scanner.stop()


    def abortScan(self):
        _LOGGER.debug("abortScan")
        self.scanAborted = True

    def hasService(self, serviceUUID) -> bool:
        _LOGGER.debug(f"hasService serviceUUID={serviceUUID}")
        return serviceUUID in self.activeClient.services

    def hasCharacteristic(self, characteristicUUID) -> bool:
        _LOGGER.debug(f"hasCharacteristic characteristicUUID={characteristicUUID}")
        return characteristicUUID in self.activeClient.characteristics

    async def writeToCharacteristic(self, serviceUUID, characteristicUUID, content):
        _LOGGER.debug(f"writeToCharacteristic serviceUUID={serviceUUID} characteristicUUID={characteristicUUID} content={content}")
        await self.is_connected_guard()
        encryptedContent = EncryptionHandler.encrypt(content, self.settings)
        payload = self._preparePayload(encryptedContent)
        await self.activeClient.client.write_gatt_char(characteristicUUID, payload, response=True)


    async def writeToCharacteristicWithoutEncryption(self, serviceUUID, characteristicUUID, content, response=True):
        _LOGGER.debug(f"writeToCharacteristicWithoutEncryption serviceUUID={serviceUUID} characteristicUUID={characteristicUUID} content={content}")
        await self.is_connected_guard()
        payload = self._preparePayload(content)
        await self.activeClient.client.write_gatt_char(characteristicUUID, payload, response=response)


    async def readCharacteristic(self, serviceUUID, characteristicUUID):
        _LOGGER.debug(f"readCharacteristic serviceUUID={serviceUUID} characteristicUUID={characteristicUUID}")
        data = await self.readCharacteristicWithoutEncryption(serviceUUID, characteristicUUID)
        if self.settings.isEncryptionEnabled():
            return EncryptionHandler.decrypt(data, self.settings)


    async def readCharacteristicWithoutEncryption(self, serviceUUID, characteristicUUID):
        _LOGGER.debug(f"readCharacteristicWithoutEncryption serviceUUID={serviceUUID} characteristicUUID={characteristicUUID}")
        await self.is_connected_guard()
        return await self.activeClient.client.read_gatt_char(characteristicUUID)


    async def setupSingleNotification(self, serviceUUID, characteristicUUID, writeCommand, timeout = None):
        if timeout is None:
            timeout = 12.5

        _LOGGER.debug(f"setupSingleNotification serviceUUID={serviceUUID} characteristicUUID={characteristicUUID}")
        await self.is_connected_guard()

        # setup the collecting of the notification data.
        _LOGGER.debug(f"setupSingleNotification: subscribe for notifications.")
        notificationDelegate = NotificationDelegate(self._killNotificationLoop, self.settings)
        await self.activeClient.subscribeNotifications(characteristicUUID, notificationDelegate.handleNotification)

        # execute something that will trigger the notifications
        _LOGGER.debug(f"setupSingleNotification: writeCommand().")
        await writeCommand()

        # wait for the results to come in.
        self.notificationLoopActive = True
        loopCount = 0
        pollInterval = 0.1
        while self.notificationLoopActive and loopCount < (timeout / pollInterval):
            await asyncio.sleep(pollInterval)
            loopCount += 1


        if notificationDelegate.result is None:
            if self.activeClient is not None:
                self.activeClient.unsubscribeNotifications(characteristicUUID)
            raise CrownstoneBleException(BleError.NO_NOTIFICATION_DATA_RECEIVED, "No notification data received.")

        if self.activeClient is not None:
            self.activeClient.unsubscribeNotifications(characteristicUUID)
        return notificationDelegate.result


    async def setupNotificationStream(self, serviceUUID, characteristicUUID, writeCommand, resultHandler, timeout):
        _LOGGER.debug(f"setupNotificationStream serviceUUID={serviceUUID} characteristicUUID={characteristicUUID} timeout={timeout}")
        await self.is_connected_guard()

        # Collect all merged notifications in a queue, so we can process them all.
        mergedNotifications = []
        def onMergedNotification():
            mergedNotifications.append(notificationDelegate.result)
            notificationDelegate.reset()

        # Subscribe for notifications and let the delegate merge them.
        _LOGGER.debug(f"setupNotificationStream: subscribe for notifications.")
        notificationDelegate = NotificationDelegate(onMergedNotification, self.settings)
        await self.activeClient.subscribeNotifications(characteristicUUID, notificationDelegate.handleNotification)

        # Execute something that will trigger the notifications.
        _LOGGER.debug(f"setupNotificationStream: writeCommand().")
        await writeCommand()

        # Wait for the results to come in.
        self.notificationLoopActive = True
        loopCount = 0
        successful = False
        pollInterval = 0.1
        while self.notificationLoopActive and loopCount < (timeout / pollInterval):
            await asyncio.sleep(pollInterval)
            _LOGGER.debug(f"loopActive={self.notificationLoopActive} loopCount={loopCount}")
            loopCount += 1
            if len(mergedNotifications):
                command = resultHandler(mergedNotifications.pop(0))
                if command == ProcessType.ABORT_ERROR:
                    _LOGGER.debug("abort")
                    self.notificationLoopActive = False
                    self.activeClient.unsubscribeNotifications(characteristicUUID)
                    raise CrownstoneBleException(BleError.ABORT_NOTIFICATION_STREAM_W_ERROR, "Aborting the notification stream because the resultHandler raised an error.")
                elif command == ProcessType.FINISHED:
                    _LOGGER.debug("finished")
                    self.notificationLoopActive = False
                    successful = True
                elif command == ProcessType.CONTINUE:
                    _LOGGER.debug("continue")

        if not successful:
            self.activeClient.unsubscribeNotifications(characteristicUUID)
            raise CrownstoneBleException(BleError.NOTIFICATION_STREAM_TIMEOUT, "Notification stream not finished within timeout.")

        # remove subscription from this characteristic
        self.activeClient.unsubscribeNotifications(characteristicUUID)

    def _killNotificationLoop(self):
        _LOGGER.debug("_killNotificationLoop")
        self.notificationLoopActive = False


    def _preparePayload(self, data: list or bytes or bytearray):
        return bytearray(data)







from crownstone_ble.core.container.ScanDataUtil import fillScanDataFromAdvertisement
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.core.modules.StoneAdvertisementTracker import StoneAdvertisementTracker
from crownstone_ble.topics.BleTopics import BleTopics
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics

"""
Class that validates advertisements from topic 'SystemBleTopics.rawAdvertisementClass'.

Each MAC address will have its own 'StoneAdvertisementTracker'.

On each 'SystemBleTopics.rawAdvertisementClass', this class will:
- Call 'update()' on the StoneAdvertisementTracker of that MAC address.
- Emit 'BleTopics.advertisement', if the address is validated.
- Emit 'BleTopics.newDataAvailable' if the address is validated, and the rawAdvertisement has service data.
- Emit 'BleTopics.rawAdvertisement' for all incoming Crownstone messages.

The threading part is removed. This adds a little overhead since the cleanup is called every checkAdvertisement. On the other hand, not threading is usually no issues.
"""
class Validator:

    def __init__(self):
        BleEventBus.subscribe(SystemBleTopics.rawAdvertisementClass, self.checkAdvertisement)
        self.trackedCrownstones = {}


    def cleanupExpiredTrackers(self):
        allKeys = []
        # we first collect keys because we might delete items from this list during ticks
        for key, trackedStone in self.trackedCrownstones.items():
            allKeys.append(key)

        for key in allKeys:
            self.trackedCrownstones[key].checkForCleanup()


    def removeStone(self, address):
        del self.trackedCrownstones[address]


    def checkAdvertisement(self, advertisement):
        self.cleanupExpiredTrackers()

        if advertisement.address not in self.trackedCrownstones:
            self.trackedCrownstones[advertisement.address] = StoneAdvertisementTracker(lambda: self.removeStone(advertisement.address))

        self.trackedCrownstones[advertisement.address].update(advertisement)

        # forward all scans over this topic. It is located here instead of the delegates so it would be easier to convert the json to classes.
        data = fillScanDataFromAdvertisement(advertisement, self.trackedCrownstones[advertisement.address].verified)
        BleEventBus.emit(BleTopics.rawAdvertisement, data)
        if self.trackedCrownstones[advertisement.address].verified:
            BleEventBus.emit(BleTopics.advertisement, data)

            if not self.trackedCrownstones[advertisement.address].duplicate:
                BleEventBus.emit(BleTopics.newDataAvailable, data)



from crownstone_core.Enums import CrownstoneOperationMode
from crownstone_core.protocol.Services import DFU_ADVERTISEMENT_SERVICE_UUID

from crownstone_core.packets.Advertisement import Advertisement

from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics

SERVICE_DATA_ADTYPE = 22
NAME_ADTYPE         = 8
FLAGS_ADTYPE        = 1

class BleakScanDelegate:

    def __init__(self, settings):
        self.settings = settings

    def handleDiscovery(self, device, advertisement_data):
        serviceData = advertisement_data.service_data
        for serviceUUID, serviceData in serviceData.items():
            longUUID = serviceUUID
            if "0000c001-0000-1000-8000-00805f9b34fb" in longUUID:
                shortUUID = int(longUUID[4:8], 16)
                self.parsePayload(device.address, device.rssi, device.name, list(serviceData), shortUUID)
            elif DFU_ADVERTISEMENT_SERVICE_UUID in longUUID:
                self.parsePayload(device.address, device.rssi, device.name, list(serviceData), DFU_ADVERTISEMENT_SERVICE_UUID)


    def parsePayload(self, address, rssi, nameText, serviceDataArray, serviceUUID):
        advertisement = Advertisement(address, rssi, nameText, serviceDataArray, serviceUUID)
        if advertisement.isCrownstoneFamily():
            if advertisement.operationMode == CrownstoneOperationMode.SETUP:
                advertisement.parse()
                BleEventBus.emit(SystemBleTopics.rawAdvertisementClass, advertisement)
            else:
                try:
                    advertisement.parse(self.settings.serviceDataKey)
                except:
                    # fail silently. If we can't parse this, we just to propagate this message
                    pass

                BleEventBus.emit(SystemBleTopics.rawAdvertisementClass, advertisement)




class SystemBleTopics:
    rawAdvertisementClass = "rawAdvertisementClass"  # used to propagate verified state messages through the system
    abortScanning         = "abortScanning"  # used to propagate verified state messages through the system
    forcedDisconnect      = "forcedDisconnect"


import logging
import time

from crownstone_core.Enums import CrownstoneOperationMode
from crownstone_core.packets.Advertisement import Advertisement
from crownstone_core.packets.ServiceData import ServiceData
from crownstone_core.packets.serviceDataParsers.containers.elements.AdvTypes import AdvType

_LOGGER = logging.getLogger(__name__)

AMOUNT_OF_REQUIRED_MATCHES = 2

"""
Class that validates advertisements from a single MAC address.

Each 'update()':
- Checks if the advertisement can be validated, and sets .verified = True if so.
"""
class StoneAdvertisementTracker:

    def __init__(self, cleanupCallback):
        self.address           = None
        self.uniqueIdentifier = None
        self.crownstoneId     = 0
        self.cleanupCallback  = None
        self.verified         = False
        self.duplicate        = False

        # config
        self.timeoutDuration = 10  # seconds
        self.consecutiveMatches = 0
        
        self.timeoutTime = time.time() + self.timeoutDuration

        self.cleanupCallback = cleanupCallback


    def checkForCleanup(self):
        now = time.time()
        # check time in self.timeoutTime with current time
        if self.timeoutTime <= now:
            _LOGGER.debug(f"Timeout {self.address}")
            self.cleanupCallback()


    def update(self, advertisement: Advertisement):
        self.address = advertisement.address

        if advertisement.isCrownstoneFamily():
            self.handlePayload(advertisement)


    def handlePayload(self, advertisement: Advertisement):
        if advertisement.operationMode == CrownstoneOperationMode.DFU:
            self.verified = True
            self.consecutiveMatches = 0
        elif advertisement.operationMode == CrownstoneOperationMode.SETUP:
            self.verified = True
            self.consecutiveMatches = 0
        else:
            self.verify(advertisement.serviceData)

        self.timeoutTime = time.time() + self.timeoutDuration

        if hasattr(advertisement.serviceData.payload, "uniqueIdentifier"):
            self.uniqueIdentifier = advertisement.serviceData.payload.uniqueIdentifier



    # check if we consistently get the ID of this crownstone.
    def verify(self, serviceData):
        if not hasattr(serviceData.payload, "uniqueIdentifier"):
            return

        if not serviceData.decrypted:
            _LOGGER.debug(f"Invalidate {self.address}, decrypted={serviceData.decrypted}")
            self.invalidateDevice(serviceData)
            return

        if not hasattr(serviceData.payload, "validation") or not hasattr(serviceData.payload, "crownstoneId"):
            return

        _LOGGER.debug(f"Check {self.address}"
                      f", id={serviceData.payload.crownstoneId}"
                      f", uniqueIdentifier={serviceData.payload.uniqueIdentifier}"
                      f", validation={serviceData.payload.validation}"
                      f", opCode={serviceData.opCode}"
                      f", advType={serviceData.payload.type}")

        if self.uniqueIdentifier == serviceData.payload.uniqueIdentifier:
            self.duplicate = True
            return

        self.duplicate = False

        if serviceData.opCode == 7:
            if serviceData.payload.validation == 0xFA:
                if serviceData.payload.type == AdvType.EXTERNAL_STATE or serviceData.payload.type == AdvType.EXTERNAL_ERROR:
                    # this is an external state payload, this means that the crownstoneId does not belong to the crownstone that broadcast it.
                    return

                if serviceData.payload.crownstoneId == self.crownstoneId:
                    self.addValidMeasurement(serviceData)
                else:
                    self.invalidateDevice(serviceData)
            else:
                self.invalidateDevice(serviceData)

    def addValidMeasurement(self, serviceData: ServiceData):
        _LOGGER.debug(f"addValidMeasurement {self.address}")
        if self.consecutiveMatches >= AMOUNT_OF_REQUIRED_MATCHES:
            self.verified = True
            self.consecutiveMatches = 0
        else:
            self.consecutiveMatches += 1
        
        self.crownstoneId = serviceData.payload.crownstoneId


    def invalidateDevice(self, serviceData: ServiceData):
        if serviceData.payload.type != AdvType.EXTERNAL_STATE and serviceData.payload.type != AdvType.EXTERNAL_ERROR:
            self.crownstoneId = serviceData.payload.crownstoneId
        
        self.consecutiveMatches = 0
        self.verified = False



from crownstone_core.Enums import CrownstoneOperationMode

from crownstone_ble.core.container.ScanData import ScanData
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.topics.SystemBleTopics import SystemBleTopics


class ModeChecker:

    def __init__(self, address: str, targetMode: CrownstoneOperationMode or None, waitUntilInTargetMode=False):
        self.address = address.lower()
        self.result = None
        self.targetMode = targetMode
        self.waitUntilInTargetMode = waitUntilInTargetMode

    def handleAdvertisement(self, scanData: ScanData):
        if scanData.address != self.address:
            return

        self.result = scanData.operationMode

        if self.targetMode is not None and self.result != self.targetMode and self.waitUntilInTargetMode:
            pass
        elif self.targetMode is None and self.result == CrownstoneOperationMode.UNKNOWN:
            # if we're looking for a mode, we'll wait for the duration of the timeout in the hope it will be something other than unknown
            pass
        else:
            BleEventBus.emit(SystemBleTopics.abortScanning, True)

    def getResult(self):
        return self.result




import logging

from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.packets.microapp.MicroappInfoPacket import MicroappInfoPacket
from crownstone_core.packets.microapp.MicroappMessagePacket import MicroappMessagePacket
from crownstone_core.packets.microapp.MicroappUploadPacket import MicroappUploadPacket
from crownstone_core.protocol.BlePackets import ControlPacket
from crownstone_core.protocol.BluenetTypes import ControlType
from crownstone_ble.core.ble_modules.ControlHandler import ControlHandler

_LOGGER = logging.getLogger(__name__)


class MicroappHandler:
    def __init__(self, control: ControlHandler):
        self.control = control
        pass

    # It looks like the UART python library cannot send large commands.
    MAX_CHUNK_SIZE = 32

    async def getMicroappInfo(self) -> MicroappInfoPacket:
        resultPacket = await self.control._writeControlAndGetResult(ControlPacket(ControlType.MICROAPP_GET_INFO).serialize())
        _LOGGER.info(f"getMicroappInfo {resultPacket}")
        infoPacket = MicroappInfoPacket(resultPacket.payload)
        return infoPacket

    async def uploadMicroapp(self, data: bytearray, index: int = 0, protocol: int = 0, chunkSize: int = MAX_CHUNK_SIZE):
        chunkSize = min(chunkSize, MicroappHandler.MAX_CHUNK_SIZE)

        for i in range(0, len(data), chunkSize):
            chunk = data[i : i + chunkSize]
            # Pad the chunk with 0xFF, so the size is a multiple of 4.
            if len(chunk) % 4:
                if isinstance(chunk, bytes):
                    chunk = bytearray(chunk)
                chunk.extend((4 - (len(chunk) % 4)) * [0xFF])
            await self._uploadMicroappChunk(index, protocol, chunk, i)

    async def _uploadMicroappChunk(self, index: int, protocol: int, data: bytearray, offset: int):
        _LOGGER.info(f"Upload microapp chunk index={index} offset={offset} size={len(data)}")
        header = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        packet = MicroappUploadPacket(header, offset, data)
        controlPacket = ControlPacket(ControlType.MICROAPP_UPLOAD).loadByteArray(packet.serialize()).serialize()
        await self.control._writeControlAndWaitForSuccess(controlPacket)
        _LOGGER.info(f"uploaded chunk offset={offset}")

    async def validateMicroapp(self, index, protocol):
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_VALIDATE).loadByteArray(packet.serialize()).serialize()
        await self.control._writeControlAndGetResult(controlPacket)

    async def enableMicroapp(self, index, protocol):
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_ENABLE).loadByteArray(packet.serialize()).serialize()
        await self.control._writeControlAndGetResult(controlPacket)

    async def removeMicroapp(self, index, protocol):
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_REMOVE).loadByteArray(packet.serialize()).serialize()
        await self.control._writeControlAndWaitForSuccess(controlPacket)
        _LOGGER.info(f"Removed app {index}")

    async def sendMessage(self, index: int, protocol: int, data: bytearray):
        _LOGGER.info(f"Send message to microapp index={index}")
        header = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        packet = MicroappMessagePacket(header, data)
        controlPacket = ControlPacket(ControlType.MICROAPP_MESSAGE).loadByteArray(packet.serialize()).serialize()
        await self.control._writeControlAndWaitForSuccess(controlPacket)



from crownstone_core.Enums import CrownstoneOperationMode
from crownstone_ble.core.container.ScanData import ScanData


class Gatherer:
    
    def __init__(self):
        self.deviceList = {}
        
    def handleAdvertisement(self, scanData: ScanData):
        rssi = float(scanData.rssi)
        if float(scanData.rssi) >= 0:
            rssi = None
        
        if scanData.address not in self.deviceList:
            self.deviceList[scanData.address] = {"address": scanData.address.lower(), "setupMode": None, "validated": scanData.validated, "rssi": rssi}

        self.deviceList[scanData.address]["validated"] = True
        self.deviceList[scanData.address]["setupMode"] = scanData.operationMode == CrownstoneOperationMode.SETUP
        
        if self.deviceList[scanData.address]["rssi"] is None:
            self.deviceList[scanData.address]["rssi"] = rssi
        else:
            self.deviceList[scanData.address]["rssi"]  = 0.9 * self.deviceList[scanData.address]["rssi"] + 0.1 * rssi
        
    
    def getCollection(self):
        collectionArray = []
        for address, device in self.deviceList.items():
            collectionArray.append(device)
            
        return collectionArray




import logging

from crownstone_core.Enums import CrownstoneOperationMode

from crownstone_ble.core.ble_modules.DevHandler import DevHandler
from crownstone_ble.core.ble_modules.MicroappHandler import MicroappHandler
from crownstone_ble.core.container.ScanData import ScanData

from crownstone_ble.core.ble_modules.BleHandler import BleHandler
from crownstone_ble.core.modules.ModeChecker import ModeChecker
from crownstone_ble.topics.BleTopics import BleTopics
from crownstone_core.Exceptions import CrownstoneError, CrownstoneBleException, CrownstoneException
from crownstone_core.core.modules.EncryptionSettings import EncryptionSettings
from crownstone_core.util.JsonFileStore import JsonFileStore

from crownstone_ble.Exceptions import BleError
from crownstone_ble.core.BleEventBus import BleEventBus
from crownstone_ble.core.ble_modules.ControlHandler import ControlHandler
from crownstone_ble.core.ble_modules.SetupHandler import SetupHandler
from crownstone_ble.core.ble_modules.StateHandler import StateHandler
from crownstone_ble.core.ble_modules.DebugHandler import DebugHandler
from crownstone_ble.core.modules.Gatherer import Gatherer
from crownstone_ble.core.modules.NearestSelector import NearestSelector
from crownstone_ble.core.modules.RssiChecker import RssiChecker

_LOGGER = logging.getLogger(__name__)

class CrownstoneBle:
    __version__ = "2.5.0"
    
    def __init__(self, bleAdapterAddress: str = None):
        # bleAdapterAddress is the MAC address of the adapter you want to use.
        self.settings = EncryptionSettings()
        self.control  = ControlHandler(self)
        self.setup    = SetupHandler(self)
        self.state    = StateHandler(self)
        self.debug    = DebugHandler(self)
        self.microapp = MicroappHandler(self.control)
        self._dev     = DevHandler(self)
        self.ble      = BleHandler(self.settings, bleAdapterAddress)

        self.defaultKeysOverridden = False

        # load default keys so the lib won't crash if you don't use keys.
        self.settings.loadKeys("adminKeyForCrown",
                               "memberKeyForHome",
                               "basicKeyForOther",
                               "MyServiceDataKey",
                               "aLocalizationKey",
                               "MyGoodMeshAppKey",
                               "MyGoodMeshNetKey")

    async def shutDown(self):
        """
        Shut down the library nicely.
        """
        await self.ble.shutDown()
    
    def setSettings(self, adminKey, memberKey, basicKey, serviceDataKey, localizationKey, meshApplicationKey, meshNetworkKey):
        self.settings.loadKeys(adminKey, memberKey, basicKey, serviceDataKey, localizationKey, meshApplicationKey, meshNetworkKey)
        self.defaultKeysOverridden = True

    def loadSettingsFromDictionary(self, data):
        """
        TODO(2021-111-15, Arend): rename function to loadKeysFromDictionary
        """
        if "admin" not in data:
            raise CrownstoneBleException(CrownstoneError.ADMIN_KEY_REQUIRED)
        if "member" not in data:
            raise CrownstoneBleException(CrownstoneError.MEMBER_KEY_REQUIRED)
        if "basic" not in data:
            raise CrownstoneBleException(CrownstoneError.BASIC_KEY_REQUIRED)
        if "serviceDataKey" not in data:
            raise CrownstoneBleException(CrownstoneError.SERVICE_DATA_KEY_REQUIRED)
        if "localizationKey" not in data:
            raise CrownstoneBleException(CrownstoneError.LOCALIZATION_KEY_REQUIRED)
        if "meshApplicationKey" not in data:
            raise CrownstoneBleException(CrownstoneError.MESH_APP_KEY)
        if "meshNetworkKey" not in data:
            raise CrownstoneBleException(CrownstoneError.MESH_NETWORK_KEY)

        self.setSettings(data["admin"], data["member"], data["basic"], data["serviceDataKey"], data["localizationKey"],
                         data["meshApplicationKey"], data["meshNetworkKey"])

    def loadSettingsFromFile(self, path):
        """
        TODO(2021-111-15, Arend): rename function to loadKeysFromFile
        """
        fileReader = JsonFileStore(path)
        data = fileReader.getData()
        self.loadSettingsFromDictionary(data)


    async def connect(self, address: str, timeout: int = 5, attempts: int = 3, ignoreEncryption=False):
        """
        Connect to a Crownstone.

        :param address:           MAC address of the Crownstone, in the format: 12:34:56:78:ab:cd
        :param timeout:           Time in seconds before giving up, for each connection attempt.
        :param attempts:          Number of connection attempts.
        :param ignoreEncryption:  True when encryption will not be used for this session.
        """
        await self.ble.connect(address, timeout=timeout, attempts=attempts)
        if not ignoreEncryption:
            await self.control._getAndSetSessionNonce()

    async def setupCrownstone(self, address, sphereId, crownstoneId, meshDeviceKey, ibeaconUUID, ibeaconMajor, ibeaconMinor):
        if not self.defaultKeysOverridden:
            raise CrownstoneBleException(BleError.NO_ENCRYPTION_KEYS_SET,
                                         "Keys are not initialized so I can't put anything on the Crownstone. "
                                         "Make sure you call .setSettings, loadSettingsFromFile or loadSettingsFromDictionary")

        await self.setup.setup(address, sphereId, crownstoneId, meshDeviceKey, ibeaconUUID, ibeaconMajor, ibeaconMinor)

    async def disconnect(self):
        await self.ble.disconnect()
    
    async def startScanning(self, scanDuration=3):
        await self.ble.scan(scanDuration)

    async def stopScanning(self):
        await self.ble.stopScanning()


    async def getCrownstonesByScanning(self, scanDuration=3):
        gatherer = Gatherer()
        subscriptionIdAll = BleEventBus.subscribe(BleTopics.rawAdvertisement, lambda scanData: gatherer.handleAdvertisement(scanData))
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionIdAll)
        return gatherer.getCollection()


    async def isCrownstoneInSetupMode(self, address: str, scanDuration=3, waitUntilInSetupMode=False) -> bool:
        _LOGGER.warning("isCrownstoneInSetupMode is deprecated. Will be removed in v3. Use either getMode or waitForMode instead.")
        """
        This will wait until it has received an advertisement from the Crownstone with the specified address. Once it has received an advertisement, it knows the mode.
        With default value for waitUntilInSetupMode (False), it will return True if the Crownstone is in setup mode, False if it isn't.

        You can use the boolean waitUntilInSetupMode to have it ignore advertisements from this Crownstone in other modes than setup mode.

        It can throw the following CrownstoneBleException
        - BleError.NO_SCANS_RECEIVED
            We have not received any scans from this Crownstone, and can't say anything about it's state.
        """
        _LOGGER.debug(f"isCrownstoneInSetupMode address={address} scanDuration={scanDuration} waitUntilInSetupMode={waitUntilInSetupMode}")
        checker = ModeChecker(address, CrownstoneOperationMode.SETUP, waitUntilInSetupMode)
        subscriptionId = BleEventBus.subscribe(BleTopics.advertisement, checker.handleAdvertisement)
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionId)
        result = checker.getResult()

        if result is None:
            raise CrownstoneBleException(BleError.NO_SCANS_RECEIVED, f'During the {scanDuration} seconds of scanning, no advertisement was received from this address.')

        return result


    async def isCrownstoneInNormalMode(self, address, scanDuration=3, waitUntilInNormalMode=False) -> bool:
        _LOGGER.warning("isCrownstoneInNormalMode is deprecated. Will be removed in v3. Use either getMode or waitForMode instead.")
        """
        This will wait until it has received an advertisement from the Crownstone with the specified address. Once it has received an advertisement, it knows the mode.
        With default value for waitUntilInSetupMode (False), it will return True if the Crownstone is in normal mode, False if it isn't.

        You can use the boolean waitUntilInNormalMode, to have it ignore advertisements from this Crownstone in other modes than setup mode.

        It can throw the following CrownstoneBleException
        - BleError.NO_SCANS_RECEIVED
            We have not received any scans from this Crownstone, and can't say anything about it's state.
        """
        _LOGGER.debug(f"isCrownstoneInNormalMode address={address} scanDuration={scanDuration} waitUntilInRequiredMode={waitUntilInNormalMode}")
        checker = ModeChecker(address, CrownstoneOperationMode.NORMAL, waitUntilInNormalMode)
        subscriptionId = BleEventBus.subscribe(BleTopics.rawAdvertisement, lambda scanData: checker.handleAdvertisement(scanData))
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionId)
        result = checker.getResult()

        if result is None:
            raise CrownstoneBleException(BleError.NO_SCANS_RECEIVED, f'During the {scanDuration} seconds of scanning, no advertisement was received from this address.')

        return result



    async def getMode(self, address, scanDuration=3) -> CrownstoneOperationMode:
        """
        Get the operation mode of the Crownstone with given MAC address.
        This will scan for advertisements, and return on the first useful advertisement from the given MAC address.
        :param address:                The MAC address of the Crownstone.
        :param scanDuration:           Timeout in seconds.
        :returns:                      The operation mode of the Crownstone.
        :raises BleError.NO_SCANS_RECEIVED: On timeout, no useful advertisements have been received.
        """
        _LOGGER.debug(f"getMode address={address} scanDuration={scanDuration}")
        checker = ModeChecker(address, None)
        subscriptionId = BleEventBus.subscribe(BleTopics.rawAdvertisement, lambda scanData: checker.handleAdvertisement(scanData))
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionId)
        result = checker.getResult()

        if result is None:
            raise CrownstoneBleException(BleError.NO_SCANS_RECEIVED, f'During the {scanDuration} seconds of scanning, no advertisement was received from this address.')

        return result


    async def waitForMode(self, address, requiredMode: CrownstoneOperationMode, scanDuration=5):
        """
        This will wait until it has received an advertisement from the Crownstone with the specified address. Once it has received an advertisement, it knows the mode. We will
        scan for the scanDuration amount of seconds or until the Crownstone is in the required mode.

        It can throw the following CrownstoneBleException
        - BleError.NO_SCANS_RECEIVED
            We have not received any scans from this Crownstone, and can't say anything about it's state.
        - BleError.DIFFERENT_MODE_THAN_REQUIRED
            During the {scanDuration} seconds of scanning, the Crownstone was not in the required mode.
        """
        _LOGGER.debug(f"waitForMode address={address} requiredMode={requiredMode} scanDuration={scanDuration}")
        checker = ModeChecker(address, requiredMode, True)
        subscriptionId = BleEventBus.subscribe(BleTopics.rawAdvertisement, lambda scanData: checker.handleAdvertisement(scanData))
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionId)
        result = checker.getResult()

        if result is None:
            raise CrownstoneBleException(BleError.NO_SCANS_RECEIVED, f'During the {scanDuration} seconds of scanning, no advertisement was received from this address.')
        if result != requiredMode:
            raise CrownstoneBleException(BleError.DIFFERENT_MODE_THAN_REQUIRED, f'During the {scanDuration} seconds of scanning, the Crownstone was not in the required mode..')



    async def getRssiAverage(self, address, scanDuration=3):
        checker = RssiChecker(address)
        subscriptionId = BleEventBus.subscribe(BleTopics.rawAdvertisement, lambda scanData: checker.handleAdvertisement(scanData))
        await self.ble.scan(duration=scanDuration)
        BleEventBus.unsubscribe(subscriptionId)
        return checker.getResult()


    async def getNearestCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=None) -> ScanData or None:
        return await self._getNearest(False, rssiAtLeast, scanDuration, returnFirstAcceptable, False, addressesToExclude)
    
    
    async def getNearestValidatedCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=None) -> ScanData or None:
        return await self._getNearest(False, rssiAtLeast, scanDuration, returnFirstAcceptable, True, addressesToExclude)
    
    
    async def getNearestSetupCrownstone(self, rssiAtLeast=-100, scanDuration=3, returnFirstAcceptable=False, addressesToExclude=None) -> ScanData or None:
        return await self._getNearest(True, rssiAtLeast, scanDuration, returnFirstAcceptable, True, addressesToExclude)


    async def _getNearest(self, setup, rssiAtLeast, scanDuration, returnFirstAcceptable, validated, addressesToExclude) -> ScanData or None:
        addressesToExcludeSet = set()
        if addressesToExclude is not None:
            for data in addressesToExclude:
                if hasattr(data,'address'):
                    addressesToExcludeSet.add(data.address.lower())
                elif isinstance(data, dict):
                    if "address" in data:
                        addressesToExcludeSet.add(data["address"].lower())
                    else:
                        raise CrownstoneException(CrownstoneError.INVALID_ADDRESS,
                                                  "Addresses to Exclude is either an array of addresses (like 'f7:19:a4:ef:ea:f6') or an array of dicts with the field 'address'")
                else:
                    addressesToExcludeSet.add(data.lower())

        selector = NearestSelector(setup, rssiAtLeast, returnFirstAcceptable, addressesToExcludeSet)

        topic = BleTopics.advertisement
        if not validated:
            topic = BleTopics.rawAdvertisement

        subscriptionId = BleEventBus.subscribe(topic, lambda scanData: selector.handleAdvertisement(scanData))

        await self.ble.scan(duration=scanDuration)
    
        BleEventBus.unsubscribe(subscriptionId)
        
        return selector.getNearest()



import logging

from crownstone_core import Conversion
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.packets.microapp.MicroappHeaderPacket import MicroappHeaderPacket
from crownstone_core.packets.microapp.MicroappInfoPacket import MicroappInfoPacket
from crownstone_core.packets.microapp.MicroappUploadPacket import MicroappUploadPacket
from crownstone_core.protocol.BlePackets import ControlStateGetPacket, ControlStateSetPacket, ControlPacket
from crownstone_core.protocol.BluenetTypes import StateType, ControlType

_LOGGER = logging.getLogger(__name__)

class DevHandler:
    def __init__(self, bluetoothCore):
        self.core = bluetoothCore
        
    async def setCurrentThresholdDimmer(self, currentAmp: float):
        packet = ControlStateSetPacket(StateType.CURRENT_CONSUMPTION_THRESHOLD_DIMMER)
        packet.loadUInt16(currentAmp * 1000)
        await self.core.state._setState(packet)

    async def getCurrentThresholdDimmer(self) -> float:
        rawState = await self.core.state._getState(StateType.CURRENT_CONSUMPTION_THRESHOLD_DIMMER)
        currentThresholdMilliAmp = Conversion.uint8_array_to_uint16(rawState)
        return currentThresholdMilliAmp / 1000.0

    async def getMicroappInfo(self) -> MicroappInfoPacket:
        _LOGGER.warning(f"getMicroappInfo moved to the MicroappHandler")
        resultPacket = await self.core.control._writeControlAndGetResult(ControlPacket(ControlType.MICROAPP_GET_INFO).serialize())
        _LOGGER.info(f"getMicroappInfo {resultPacket}")
        infoPacket = MicroappInfoPacket(resultPacket.payload)
        return infoPacket

    async def uploadMicroapp(self, data: bytearray, index: int = 0, protocol: int = 0, chunkSize: int = 128):
        _LOGGER.warning(f"uploadMicroapp moved to the MicroappHandler")
        for i in range(0, len(data), chunkSize):
            chunk = data[i : i + chunkSize]
            # Pad the chunk with 0xFF, so the size is a multiple of 4.
            if len(chunk) % 4:
                if isinstance(chunk, bytes):
                    chunk = bytearray(chunk)
                chunk.extend((4 - (len(chunk) % 4)) * [0xFF])
            await self.uploadMicroappChunk(index, protocol, chunk, i)

    async def uploadMicroappChunk(self, index: int, protocol: int, data: bytearray, offset: int):
        _LOGGER.warning(f"uploadMicroappChunk moved to the MicroappHandler")
        _LOGGER.info(f"Upload microapp chunk index={index} offset={offset} size={len(data)}")
        header = MicroappHeaderPacket(appIndex=index, protocol=protocol)
        packet = MicroappUploadPacket(header, offset, data)
        controlPacket = ControlPacket(ControlType.MICROAPP_UPLOAD).loadByteArray(packet.serialize()).serialize()
        await self.core.control._writeControlAndWaitForSuccess(controlPacket)
        _LOGGER.info(f"uploaded chunk offset={offset}")
        # TODO: return the final result?

    async def validateMicroapp(self, index, protocol):
        _LOGGER.warning(f"validateMicroapp moved to the MicroappHandler")
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_VALIDATE).loadByteArray(packet.serialize()).serialize()
        await self.core.control._writeControlAndGetResult(controlPacket)

    async def enableMicroapp(self, index, protocol):
        _LOGGER.warning(f"enableMicroapp moved to the MicroappHandler")
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_ENABLE).loadByteArray(packet.serialize()).serialize()
        await self.core.control._writeControlAndGetResult(controlPacket)

    async def removeMicroapp(self, index, protocol):
        _LOGGER.warning(f"removeMicroapp moved to the MicroappHandler")
        packet = MicroappHeaderPacket(index, protocol)
        controlPacket = ControlPacket(ControlType.MICROAPP_REMOVE).loadByteArray(packet.serialize()).serialize()
        await self.core.control._writeControlAndWaitForSuccess(controlPacket)
        _LOGGER.info(f"Removed app {index}")



class ScanData:

    def __init__(self):
        self.address       = None
        self.rssi          = None
        self.name          = None
        self.operationMode = None
        self.deviceType    = None
        self.payload       = None
        self.validated     = None

    def __str__(self):
        return \
           f"address:       {self.address              }\n" \
           f"rssi:          {self.rssi                 }\n" \
           f"name:          {self.name                 }\n" \
           f"operationMode: {self.operationMode        }\n" \
           f"deviceType:    {self.deviceType.__str__() }\n" \
           f"payload:       {self.payload              }\n" \
           f"validated:     {self.validated            }\n"



from crownstone_core.Exceptions import CrownstoneError, CrownstoneException
from crownstone_core.packets.ResultPacket import ResultPacket
from crownstone_core.packets.debug.AdcChannelSwapsPacket import AdcChannelSwapsPacket
from crownstone_core.packets.debug.AdcRestartsPacket import AdcRestartsPacket
from crownstone_core.packets.debug.BootloaderInfoPacket import BootloaderInfoPacket, BOOTLOADER_INFO_PROTOCOL
from crownstone_core.packets.debug.PowerSamplesPacket import PowerSamplesPacket
from crownstone_core.packets.debug.SwitchHistoryPacket import SwitchHistoryListPacket
from crownstone_core.protocol.BluenetTypes import ResultValue
from crownstone_core.protocol.Characteristics import CrownstoneCharacteristics, DeviceCharacteristics
from crownstone_core.protocol.ControlPackets import ControlPacket, ControlType
from crownstone_core.protocol.ControlPackets import ControlPacketsGenerator
from crownstone_core.protocol.Services import CSServices
from crownstone_core.util.Conversion import Conversion


class DebugHandler:
	def __init__(self, bluetoothCore):
		self.core = bluetoothCore

	async def getHardwareVersion(self) -> str:
		""" Get the hardware version of the Crownstone as string. """
		buf = await self.core.ble.readCharacteristicWithoutEncryption(CSServices.DeviceInformation, DeviceCharacteristics.HardwareRevision)
		return Conversion.uint8_array_to_string(buf)

	async def getFirmwareVersion(self) -> str:
		""" Get the firmware version of the Crownstone as string. """
		buf = await self.core.ble.readCharacteristicWithoutEncryption(CSServices.DeviceInformation, DeviceCharacteristics.FirmwareRevision)
		return Conversion.uint8_array_to_string(buf)

	async def getBootloaderVersion(self) -> str:
		""" Get the bootloader version of the Crownstone as simple string. """
		bootloaderInfo = await self.getBootloaderInfo()
		if bootloaderInfo.protocol != BOOTLOADER_INFO_PROTOCOL:
			raise CrownstoneException(CrownstoneError.PROTOCOL_NOT_SUPPORTED)
		bootloaderVersion = f"{bootloaderInfo.major}.{bootloaderInfo.minor}.{bootloaderInfo.patch}"
		if bootloaderInfo.preReleaseVersion != 255:
			bootloaderVersion += f".{bootloaderInfo.preReleaseVersion}"
		return bootloaderVersion

	async def getBootloaderInfo(self) -> BootloaderInfoPacket:
		controlPacket = ControlPacket(ControlType.GET_BOOTLOADER_VERSION).serialize()
		result = await self._writeControlAndGetResult(controlPacket)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return BootloaderInfoPacket(result.payload)

	async def getUptime(self):
		""" Get the uptime of the crownstone in seconds. """
		controlPacket = ControlPacket(ControlType.GET_UPTIME).serialize()
		result = await self._writeControlAndGetResult(controlPacket)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return Conversion.uint8_array_to_uint32(result.payload)

	async def getAdcRestarts(self):
		"""	Get number of ADC restarts since boot. Returns an AdcRestartsPacket. """
		controlPacket = ControlPacket(ControlType.GET_ADC_RESTARTS).serialize()
		result = await self._writeControlAndGetResult(controlPacket)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return AdcRestartsPacket(result.payload)

	async def getAdcChannelSwaps(self):
		""" Get number of ADC channel swaps since boot. Returns an AdcChannelSwapsPacket. """
		controlPacket = ControlPacket(ControlType.GET_ADC_CHANNEL_SWAPS).serialize()
		result = await self._writeControlAndGetResult(controlPacket)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return AdcChannelSwapsPacket(result.payload)

	async def getSwitchHistory(self):
		""" Get the switch history. Returns a SwitchHistoryListPacket. """
		controlPacket = ControlPacket(ControlType.GET_SWITCH_HISTORY).serialize()
		result = await self._writeControlAndGetResult(controlPacket)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return SwitchHistoryListPacket(result.payload)

	async def getPowerSamples(self, samplesType):
		""" Get all power samples of the given type. Returns a list of PowerSamplesPacket. """
		allSamples = []
		index = 0
		while True:
			result = await self._getPowerSamples(samplesType, index)
			if result.resultCode == ResultValue.WRONG_PARAMETER:
				return allSamples
			elif result.resultCode == ResultValue.SUCCESS:
				samples = PowerSamplesPacket(result.payload)
				allSamples.append(samples)
				index += 1
			else:
				raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))

	async def getPowerSamplesAtIndex(self, samplesType, index):
		""" Get power samples of given type at given index. Returns a PowerSamplesPacket. """
		result = await self._getPowerSamples(samplesType, index)
		if result.resultCode != ResultValue.SUCCESS:
			raise CrownstoneException(CrownstoneError.RESULT_NOT_SUCCESS, "Result: " + str(result.resultCode))
		return PowerSamplesPacket(result.payload)

	async def _getPowerSamples(self, samplesType, index):
		""" Get power samples of given type at given index, but don't check result code. """
		controlPacket = ControlPacketsGenerator.getPowerSamplesRequestPacket(samplesType, index)
		return await self._writeControlAndGetResult(controlPacket)

	async def _writeControlPacket(self, packet):
		""" Write the control packet. """
		await self.core.control._writeControlPacket(packet)

	async def _writeControlAndGetResult(self, controlPacket) -> ResultPacket:
		return await self.core.control._writeControlAndGetResult(controlPacket)
